from math import (
    acos,
    acosh,
    asin,
    asinh,
    atan,
    atan2,
    atanh,
    ceil,
    cos,
    degrees,
    erf,
    erfc,
    exp,
    expm1,
    fabs,
    factorial,
    floor,
    fsum,
    gamma,
    isqrt,
    ldexp,
    lgamma,
    log1p,
    log2,
    log10,
    nextafter,
    prod,
    radians,
    sin,
    sqrt,
    tan,
    trunc,
    ulp,
)
from operator import invert, not_
from statistics import (
    fmean,
    geometric_mean,
    harmonic_mean,
    mean,
    median,
    median_grouped,
    median_high,
    median_low,
    mode,
    multimode,
    pstdev,
    pvariance,
    stdev,
    variance,
)


def dt(value):
    """Convert value to datetime using datetime.datetime.fromisoformat."""
    if isinstance(value, str | int):
        import datetime

        return datetime.datetime.fromisoformat(f'{value}')

    import jdatetime

    if isinstance(value, jdatetime.datetime):
        return value.togregorian()

    raise ValueError('unknown format for jdt')


def weeks(val):
    import datetime

    return datetime.timedelta(weeks=val)


def days(val):
    import datetime

    return datetime.timedelta(val)


def hours(val):
    import datetime

    return datetime.timedelta(hours=val)


def minutes(val):
    import datetime

    return datetime.timedelta(minutes=val)


def seconds(val):
    import datetime

    return datetime.timedelta(seconds=val)


def jdt(value):
    """Convert value to jdatetime.datetime."""
    import jdatetime

    if isinstance(value, (str, int)):
        return jdatetime.datetime.fromisoformat(f'{value}')

    import datetime

    if isinstance(value, datetime.datetime):
        return jdatetime.datetime.fromgregorian(datetime=value)

    raise ValueError('unknown format for jdt')


UNARY_OPERATORS = {
    '!': factorial,
    'abs': abs,
    'acos': acos,
    'acosh': acosh,
    'asin': asin,
    'asinh': asinh,
    'atan': atan,
    'atan2': atan2,
    'atanh': atanh,
    'bin': bin,
    'bool': bool,
    'ceil': ceil,
    'chr': chr,
    'cos': cos,
    'days': days,
    'deg': degrees,
    'dt': dt,
    'erf': erf,
    'erfc': erfc,
    'exp': exp,
    'expm1': expm1,
    'fabs': fabs,
    'floor': floor,
    'fmean': fmean,
    'fsum': fsum,
    'gamma': gamma,
    'gmean': geometric_mean,
    'hex': hex,
    'hmean': harmonic_mean,
    'hours': hours,
    'isqrt': isqrt,
    'jdt': jdt,
    'ldexp': ldexp,
    'lgamma': lgamma,
    'log10': log10,
    'log1p': log1p,
    'log2': log2,
    'max': max,
    'mean': mean,
    'med': median,
    'medg': median_grouped,
    'medh': median_high,
    'medl': median_low,
    'min': min,
    'minutes': minutes,
    'mode': mode,
    'multimode': multimode,
    'nextafter': nextafter,
    'not': not_,
    'oct': oct,
    'prod': prod,
    'pstdev': pstdev,
    'pvar': pvariance,
    'rad': radians,
    'round': round,
    'seconds': seconds,
    'sin': sin,
    'sqrt': sqrt,
    'stdev': stdev,
    'sum': sum,
    'tan': tan,
    'trunc': trunc,
    'ulp': ulp,
    'var': variance,
    'weeks': weeks,
    '~': invert,
}



__version__ = '0.16.1'

import math
from math import floor, log10
from pprint import pprint
from statistics import correlation, covariance, linear_regression

from regex import compile as rc

from _kalk.binary_ops import BINARY_OPERATORS
from _kalk.unary_ops import UNARY_OPERATORS

STACK = []
STACKS = [STACK]


STORAGE = {}


def two_arg_factory(func):
    def f():
        d = func(STACK[-1], STACK[-2])
        del STACK[-1]
        STACK[-1] = d

    f.__doc__ = func.__doc__
    f.__name__ = func.__name__
    return f


def print_stack():
    for i, n in enumerate(STACK):
        print(f'{i}: {n}')


def display_help():
    pprint(BINARY_OPERATORS)
    pprint(UNARY_OPERATORS)
    pprint(SPECIAL_OPERATORS)


def load_constant_factory(name):
    val = getattr(math, name)

    def load_constant():
        STACK.append(val)

    load_constant.__doc__ = f"""Load {name} = {val} into the stack."""
    return load_constant


def ans():
    STACK.append(STACK[-1])


def swap():
    """Swap the place of last two results on the stack"""
    STACK[-2:] = reversed(STACK[-2:])


def paste_from_clipboard():
    from pyperclip import paste

    evaluate(paste())


def copy_to_clipboard():
    from pyperclip import copy

    copy(f'{STACK[-1]}')


def delete():
    del STACK[-STACK[-1] - 1 :]


def store():
    k = STACK.pop()
    STORAGE[k] = STACK.pop()


def recall():
    STACK.append(STORAGE[STACK.pop()])


PRECISION = 5


def precision():
    global PRECISION
    PRECISION = STACK.pop()
    try:
        return STACK[-1]
    except IndexError:
        return None


FORMAT = '{:,}'.format


def set_normal_format():
    global FORMAT
    FORMAT = '{:,}'.format


SI = 0


def toggle_si_format():
    global SI
    SI = not SI


def set_eng_format():
    global FORMAT

    # from https://stackoverflow.com/a/19270863/2705757 with some modifications
    def eng_string(x):
        """
        Returns float/int value <x> formatted in a simplified engineering format -
        using an exponent that is a multiple of 3.

        format: printf-style string used to format the value before the exponent.

        si: if true, use SI suffix for exponent, e.g. k instead of e3, n instead of
        e-9 etc.

        E.g. with format='%.2f':
            1.23e-08 => 12.30e-9
                 123 => 123.00
              1230.0 => 1.23e3
          -1230000.0 => -1.23e6

        and with si=True:
              1230.0 => 1.23k
          -1230000.0 => -1.23M
        """
        if x < 0:
            x = -x
            sign = '-'
        elif x > 0:
            sign = ''
        else:
            return '0'

        exp = int(floor(log10(x)))
        exp3 = exp - (exp % 3)
        x3 = x / (10**exp3)

        if SI is True and -24 <= exp3 <= 24 and exp3 != 0:
            exp3_text = 'yzafpnum kMGTPEZY'[(exp3 + 24) // 3]
        elif exp3 == 0:
            exp3_text = ''
        else:
            exp3_text = f'e{exp3}'

        return f'{sign}{x3:.{max(PRECISION, 3)}g}{exp3_text}'

    FORMAT = eng_string


def set_sci_format():
    global FORMAT

    def sci_string(n):
        return f'{n:.{PRECISION}e}'

    FORMAT = sci_string


def set_general_format():
    global FORMAT

    def sci_string(n):
        return f'{n:.{PRECISION}g}'

    FORMAT = sci_string


def call_method(identifier: str):
    STACK.append(getattr(STACK.pop(), identifier)())


def now():
    import datetime

    STACK.append(datetime.datetime.now())


def utcnow():
    import datetime

    STACK.append(datetime.datetime.utcnow())


def today():
    import datetime

    STACK.append(datetime.date.today())


def str_help():
    string = STACK.pop()
    op = (
        UNARY_OPERATORS.get(string)
        or SPECIAL_OPERATORS.get(string)
        or UNARY_OPERATORS.get(string)
    )
    help(op)


def clear_stack():
    del STACK[:]


def start_substack():
    global STACK
    new_stack = []
    STACK.append(new_stack)
    STACK = new_stack
    STACKS.append(new_stack)


def end_substack():
    global STACK
    STACKS.pop()
    STACK = STACKS[-1]


def enter_substack():
    global STACK
    STACK = STACK[-1]
    STACKS.append(STACK)


def exit_():
    raise SystemExit


SPECIAL_OPERATORS = {
    '<>': swap,
    '?': str_help,
    'SI': toggle_si_format,
    '[': start_substack,
    ']': end_substack,
    'a': ans,
    'c': clear_stack,
    'corr': two_arg_factory(correlation),
    'covar': two_arg_factory(covariance),
    'cp': copy_to_clipboard,
    'del': delete,
    'dist': two_arg_factory(math.dist),
    'eng': set_eng_format,
    'es': enter_substack,
    'exit': exit_,
    'gen': set_general_format,
    'h': display_help,
    'linreg': two_arg_factory(linear_regression),
    'now': now,
    'nrm': set_normal_format,
    'prec': precision,
    'pst': paste_from_clipboard,
    'rcl': recall,
    's': print_stack,
    'sci': set_sci_format,
    'sto': store,
    'today': today,
    'utcnow': utcnow,
}

for const in 'tau', 'pi', 'e', 'nan', 'inf':
    SPECIAL_OPERATORS[const] = load_constant_factory(const)


N = (  # noqa
    r'(?>'
    r'[\d۰-۹][\d۰-۹,_]*+(?:\.[\d۰-۹_]*+)?+'
    r'|\.[\d۰-۹_]++'
    r')(?:[Ee][+-]?+[\d۰-۹_]++)?+'
)
fullmatch = rc(  # noqa
    r'\s*+'
    r'(?:'
    r'('  # each token is either a number or an operator
    rf'[+-]?+{N}(?:[Jj]|[-+]{N}[Jj])?+'  # complex part
    r'|"[^"]*"'
    r'|\L<operators>'
    r'|\.[^\d\W]\w*'
    r')\s*+'
    r')*+',
    operators=(
        BINARY_OPERATORS.keys()
        | UNARY_OPERATORS.keys()
        | SPECIAL_OPERATORS.keys()
    ),
).fullmatch


def operate(token):
    if (op := BINARY_OPERATORS.get(token)) is not None:
        last = STACK.pop()
        try:
            STACK.append(op(STACK.pop(), last))
        except IndexError:
            STACK.append(last)
            raise
    elif (op := UNARY_OPERATORS.get(token)) is not None:
        STACK.append(op(STACK.pop()))
    elif (op := SPECIAL_OPERATORS.get(token)) is not None:
        op()
    elif token[0] == '.' and (identifier := token[1:]).isidentifier():
        call_method(identifier)
    else:
        return False


def evaluate(i):
    m = fullmatch(i)
    if m is None:
        print('SyntaxError')
        return None
    for token in m.captures(1):
        try:
            if operate(token) is not False:
                continue  # the token was an operator
        except IndexError:
            print('Error: not enough arguments')
            continue
        except Exception as e:
            print(type(e).__name__)
            return

        if token[0] == '"':
            STACK.append(eval(token))
            continue

        token = token.replace(',', '').replace('_', '')
        try:
            try:
                result = int(token)
            except ValueError:
                result = float(token)
        except ValueError:
            result = complex(token)

        STACK.append(result)

    try:
        return STACK[-1]
    except IndexError:
        return None


def main():
    print(f'Kalk v{__version__}')
    while True:
        try:
            last_result = evaluate(input('>>> '))
        except KeyboardInterrupt:
            print()
            continue
        if isinstance(last_result, (int, float)):
            print(FORMAT(last_result))
        elif last_result is not None:
            print(last_result)


if __name__ == '__main__':
    main()



from math import (
    comb,
    copysign,
    fmod,
    gcd,
    hypot,
    lcm,
    log,
    perm,
    pow as fpow,
    remainder,
)
from operator import (
    add,
    and_,
    eq,
    floordiv,
    ge,
    gt,
    is_,
    le,
    lshift,
    lt,
    mod,
    mul,
    ne,
    neg,
    or_,
    rshift,
    sub,
    truediv,
    xor,
)


def percent(a, b, /):
    return (b - a) / a * 100


BINARY_OPERATORS = {
    '!=': ne,
    '%%': percent,
    '%': mod,
    '&': and_,
    '*': mul,
    '**': pow,
    '+': add,
    '-': sub,
    '/': truediv,
    '//': floordiv,
    '<': lt,
    '<<': lshift,
    '<=': le,
    '==': eq,
    '>': gt,
    '>=': ge,
    '>>': rshift,
    'C': comb,
    'P': perm,
    '^': xor,
    'cs': copysign,
    'fmod': fmod,
    'fpow': fpow,
    'gcd': gcd,
    'hypot': hypot,
    'is': is_,
    'lcm': lcm,
    'log': log,
    'rem': remainder,
    '|': or_,
    '~': neg,
    '٪': mod,
    '٪٪': percent,
}



from unittest.mock import patch

from _kalk import STACK, evaluate


def test_add():
    assert evaluate('1 +2 +') == 3


def test_negative():
    assert evaluate('-1 -2 -') == 1


def test_empty():
    del STACK[:]
    assert evaluate('') is None
    STACK.append(1)
    assert evaluate('') == 1


def test_percent():
    assert evaluate('1 2 %%') == 100.0


def test_persian_digits():
    assert evaluate('۱ ۲ /') == 0.5


def test_float():
    assert evaluate('1.2 1e2 *') == 120


def test_cos():
    assert evaluate('pi cos') == -1.0


def test_factorial():
    assert evaluate('3!') == 6


@patch('builtins.print')
def test_invalid_op(mocked_print):
    assert evaluate('c sin') is None
    mocked_print.assert_called_once_with('Error: not enough arguments')


def test_eulers_number():
    import math

    assert evaluate('e') is math.e


def test_distance():
    assert evaluate('[0 1] [0 0] dist') == 1.0


def test_chr():
    assert evaluate('65 chr') == 'A'


def test_sum_fsum():
    assert evaluate('[[1 2 3]sum') == 6
    assert evaluate('6]fsum') == 12.0


def test_thousand_separator():
    assert evaluate('1,234 5 +') == 1239


def test_swap():
    evaluate('c 1 2 <>')
    assert STACK == [2, 1]


def test_float_without_leading_zero():
    assert evaluate('.9') == 0.9


def test_leading_positive_sign():
    assert evaluate('+1 -1 +') == 0


def test_complex():
    assert evaluate('1+2j 0j -') == 1 + 2j


def test_capital_e():
    assert evaluate('2E-2') == 0.02


@patch('builtins.print')
def test_syntax_error(mocked_print):
    assert evaluate('invalid') is None
    mocked_print.assert_called_once_with('SyntaxError')


@patch('builtins.print')
def test_preserve_stack_on_binary_fail(mocked_print):
    evaluate('c 1 *')
    assert STACK == [1]
    mocked_print.assert_called_once_with('Error: not enough arguments')


def test_ignore_underscore_in_numbers():
    assert evaluate('1_2.3_4e0_6J') == 12340000j


@patch('pyperclip.paste', lambda: '+0j')
@patch('pyperclip.copy')
def test_copy_paste(copy_mock):
    assert evaluate('pst') == 0j
    evaluate('cp')
    copy_mock.assert_called_once_with('0j')


def test_pop():
    assert evaluate('c 7 8 1 del') == 7
    assert STACK == [7]


def test_answer():
    assert evaluate('7 a +') == 14


@patch('builtins.print')
def test_sto_rcl(mocked_print):
    assert evaluate('c 1 2 sto 3 4 + 2 rcl +') == 8
    assert evaluate('1 rcl +') is None
    mocked_print.assert_called_once_with('KeyError')


@patch('builtins.print')
def test_zero_division(mocked_print):
    assert evaluate('1 0 /') is None
    mocked_print.assert_called_once_with('ZeroDivisionError')


def test_notations():
    import _kalk

    v = evaluate('1 3 /')
    assert _kalk.FORMAT(v) == '0.3333333333333333'
    v = evaluate('eng')
    assert _kalk.FORMAT(v) == '333.33e-3'  # default precision is 5
    v = evaluate('sci')
    assert _kalk.FORMAT(v) == '3.33333e-01'
    v = evaluate('gen')
    assert _kalk.FORMAT(v) == '0.33333'
    v = evaluate('nrm')
    assert _kalk.FORMAT(v) == '0.3333333333333333'
    v = evaluate('2 prec')
    assert _kalk.FORMAT(v) == '0.3333333333333333'
    v = evaluate('eng')
    assert _kalk.FORMAT(v) == '333e-3'
    assert _kalk.FORMAT(0) == '0'
    v = evaluate('sci')
    assert _kalk.FORMAT(v) == '3.33e-01'
    v = evaluate('gen')
    assert _kalk.FORMAT(v) == '0.33'


def test_as_integer_ratio_method():
    assert evaluate('.25 .as_integer_ratio') == (1, 4)
    assert evaluate('2 .as_integer_ratio') == (2, 1)


def test_strings():
    assert evaluate('"a" "b" +') == 'ab'
    assert evaluate('.upper') == 'AB'


def test_datetime_timedelta():
    from datetime import timedelta

    assert evaluate('"2023-03-22" dt "2023-02-22" dt -') == timedelta(28)
    assert evaluate('28 days +') == timedelta(56)
    assert evaluate('"1402-01-02" jdt dt "2023-03-22" dt ==') is True


def test_jdatetime():
    assert evaluate('"1402-01-02" jdt "2023-03-22" dt jdt ==') is True


def test_now():
    import datetime

    assert isinstance(evaluate('now'), datetime.datetime)


def test_lt():
    assert evaluate('1 2 <') == 1


def test_gmean():
    assert evaluate('[54 24 36] gmean round') == 36.0


def test_substacks():
    evaluate('c 1 2 [3 4] es 5 ] 6')
    assert STACK == [1, 2, [3, 4, 5], 6]
